"""
Configuration definition

This module defines the :class:`.Config` base class, from which all configuration derive.

.. rubric:: Types

This module uses the following types.

.. py:data:: _Config

    Configuration dataclass type being constructed

"""

from __future__ import annotations

import argparse
import inspect
import os
import sys
from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Callable,
    ClassVar,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    get_args,
)

from typing_extensions import Annotated, get_origin, get_type_hints

from . import types
from .arg import Param
from .processor import Processor, SpecialAction
from .userr import Err, Res

_Config = TypeVar("_Config", bound="Config")


@dataclass(frozen=True)
class IniSection:
    """
    Describes a section of an INI file to include in the current configuration
    """

    name: str  #: Section name
    strict: bool  #: Whether all the keys must correspond to parsed arguments


@dataclass(frozen=True)
class Config(ABC):
    """
    Base class for dataclasses holding configuration data

    Example:
        .. code-block:: python

            @dataclass(frozen=True)
            class Adder(Config):
                x: Annotated[int, Param.store(types.int_)]
                y: Annotated[int, Param.store(types.int_, default_value="0")]

            res = Adder.from_command_line_(args=['x','2', 'y','3'], env={})
            res.x + res.y
    """

    #: Names of sections to parse in configuration files, with unknown keys ignored
    ini_relaxed_sections_: ClassVar[Sequence[str]] = ["Common", "COMMON", "common"]

    #: Names of additional sections to parse in configuration files, unknown keys error
    ini_strict_sections_: ClassVar[Sequence[str]] = []

    @classmethod
    def ini_sections_(cls) -> Sequence[IniSection]:
        """
        Returns a sequence of INI file sections to parse

        By default, this parses first the relaxed sections and then the strict ones.

        This method can be overridden.
        """
        relaxed = [IniSection(name, False) for name in cls.ini_relaxed_sections_]
        strict = [IniSection(name, True) for name in cls.ini_strict_sections_]
        return relaxed + strict

    prog_: ClassVar[Optional[str]] = None  #: Program name
    description_: ClassVar[Optional[str]] = None  #: Text to display before the argument help
    env_prefix_: ClassVar[Optional[str]] = None  #: Prefix for environment variables

    @classmethod
    def validators_(cls: Type[_Config]) -> Sequence[Callable[[_Config], Optional[Err]]]:
        """
        Returns all validators present in the given subclass of this class

        Validators are methods that take no arguments (except self) and return an optional error.
        Their name starts with ``validate_`` and ends with an underscore.

        Returns:
            A sequence of all validators
        """
        res: List[Callable[[_Config], Optional[Err]]] = []
        for name, meth in inspect.getmembers(cls, inspect.isroutine):
            if name.startswith("validate_") and name.endswith("_"):
                res.append(getattr(cls, name))
        return res

    @classmethod
    def version_(cls) -> Optional[str]:
        """
        Returns the version number of this script

        Designed to be overridden by a subclass

        Returns:
            Version
        """
        return None

    @classmethod
    def processor_(cls: Type[_Config]) -> Processor[_Config]:
        """
        Returns a processor for this configuration
        """
        return Processor.make(cls)

    @classmethod
    def parse_command_line_(
        cls: Type[_Config],
        cwd: Optional[Path] = None,
        args: Optional[Sequence[str]] = None,
        env: Optional[Mapping[str, str]] = None,
    ) -> Res[Union[_Config, SpecialAction]]:
        """
        Parses multiple information sources, returns a configuration, a command or an error

        Default values are taken from the current working directory, the script command line
        arguments, and the current environment variables.

        Args:
            cwd: Directory used as a base for the configuration file relative paths
            args: Command line arguments
            env: Environment variables

        Returns:
            A parsed configuration or an error
        """
        if cwd is None:
            cwd = Path.cwd()
        if args is None:
            args = sys.argv[1:]
        if env is None:
            env = os.environ
        processor = cls.processor_()
        return processor.process(cwd, args, env)

    @classmethod
    def from_command_line_(
        cls: Type[_Config],
        cwd: Optional[Path] = None,
        args: Optional[Sequence[str]] = None,
        env: Optional[Mapping[str, str]] = None,
    ) -> _Config:
        """
        Parses multiple information sources into a configuration and display help on error

        Default values are taken from the current working directory, the script command line
        arguments, and the current environment variables.

        Args:
            cwd: Directory used as a base for the configuration file relative paths
            args: Command line arguments
            env: Environment variables

        Returns:
            A parsed configuration
        """
        if cwd is None:
            cwd = Path.cwd()
        if args is None:
            args = sys.argv[1:]
        if env is None:
            env = os.environ
        res = cls.parse_command_line_(cwd, args, env)

        if isinstance(res, cls):
            return res

        if isinstance(res, Err):
            print("Encountered errors:")
            res.pretty_print()
            print(" ")
            cls.get_argument_parser_().print_help()
            sys.exit(1)

        assert isinstance(res, SpecialAction)
        if res == SpecialAction.HELP:
            cls.processor_().argument_parser.print_help()
            sys.exit(0)
        elif res == SpecialAction.VERSION:
            v = cls.version_()
            if v is None:
                v = "Unknown version number"
            print(v)
            sys.exit(0)
        else:
            raise NotImplementedError(f"Unknown special action {res}")

    @classmethod
    def get_argument_parser_(cls: Type[_Config]) -> argparse.ArgumentParser:
        """
        Returns an :class:`argparse.ArgumentParser` for documentation purposes
        """
        return cls.processor_().argument_parser



"""
Argument value parsing

This module is mostly self-contained, and provides ways to construct :class:`.ParamType` instances
which parse string arguments into values.

During the configuration building, the parsed values are collected by a 
:class:`configpile.collector.Collector` instance.

.. rubric:: Types

This module uses the following types.

.. py:data:: _Value

    Value being parsed by a :class:`.ParamType`

.. py:data:: _Parameter

    Type received by a mapping function

.. py:data:: _ReturnType

    Type returned by a mapping function

.. py:data:: _Item

    Item in a sequence
"""

from __future__ import annotations

import pathlib
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    Literal,
    Mapping,
    NoReturn,
    Optional,
    Protocol,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

if TYPE_CHECKING:
    import parsy

from . import userr
from .userr import Err, Res, collect_seq
from .util import assert_never

# those types are documented in the module docstring

_I = TypeVar("_I")  # type used only internally

_Value = TypeVar("_Value")

_Item = TypeVar("_Item")

_Parameter = TypeVar("_Parameter")

_ReturnType = TypeVar("_ReturnType")


class ForceCase(Enum):
    """
    Describes whether a string is normalized to lower or upper case before processing
    """

    NO_CHANGE = 0  #: Keep case
    UPPER = 1  #: Change to uppercase
    LOWER = 2  #: Change to lowercase


class ParamType(ABC, Generic[_Value]):
    """Describes a parameter type, which parses a string into a parameter value"""

    @abstractmethod
    def parse(self, arg: str) -> Res[_Value]:
        """
        Parses a string into a result

        This method reports parsing errors using a result type instead of raising
        exceptions.

        Example:
            >>> user_input = "invalid"
            >>> float_.parse(user_input)
            Err1("Error 'could not convert string to float: 'invalid'' in 'invalid'")
            >>> user_input = 2.0
            >>> float_.parse(user_input)
            2.0

        Args:
            arg: String value to parse

        Returns:
            A result containing either the parsed value or a description of an error
        """

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        """
        Returns, if any, the keyword arguments that should be provided to :mod:`argparse`

        The arguments will be added to :meth:`argparse.ArgumentParser.add_argument`
        """
        return {}

    def map(self, f: Callable[[_Value], _Item]) -> ParamType[_Item]:
        """
        Maps successful results of the parser through the given function

        Args:
            f: Function to map the result through

        Returns:
            Updated parameter type
        """
        return _Mapped(self, f)

    def as_sequence_of_one(self) -> ParamType[Sequence[_Value]]:
        """
        Returns a parameter type, that returns a sequence of a single value on success

        Returns:
            Updated parameter type
        """
        f: Callable[[_I], Sequence[_I]] = lambda t: [t]
        return self.map(f)

    def empty_means_none(self, strip: bool = True) -> ParamType[Optional[_Value]]:
        """
        Returns a new parameter type where the empty string means None

        Args:
            strip: Whether to strip whitespace

        Returns:
            A new parameter type
        """
        return _EmptyMeansNone(self, strip)

    def separated_by(
        self, sep: str, strip: bool = True, remove_empty: bool = True
    ) -> ParamType[Sequence[_Value]]:
        """
        Returns a new parameter type that parses values separated by a string

        Args:
            sep: Separator
            strip: Whether to strip whitespace from separated values
            remove_empty: Whether to remove empty strings before parsing them

        Returns:
            A new parameter type
        """
        return _SeparatedBy(self, sep, strip, remove_empty)

    def validated(
        self, predicate: Callable[[_Value], bool], msg: Union[str, Callable[[str, _Value], str]]
    ) -> _Validated[_Value]:
        """
        Returns a parameter type that verifies a predicate after successful parse

        Args:
            predicate: Predicate to check
            msg: Either a string, or a function that constructs a string from the parsed string
                 and the result

        Returns:
            A new parameter type
        """
        if isinstance(msg, str):
            c: str = msg
            return _Validated(self, predicate, lambda a, t: c)
        else:
            return _Validated(
                self,
                predicate,
                msg,
            )

    @staticmethod
    def from_parser(type_: Type[_Value], parser: parsy.Parser) -> ParamType[_Value]:
        """
        Creates a parameter type from a parsy parser

        Args:
            type_: PEP 484 type, used to type the return argument
            parser: Parser returning a value of type ``t``

        Returns:
            Parameter type
        """
        return _Parsy(parser)

    @staticmethod
    def from_function_that_raises(f: Callable[[str], _Value]) -> ParamType[_Value]:
        """
        Creates a parameter type from a function that raises exceptions on parse errors

        Args:
            f: Function that parses the string

        Returns:
            Parameter type
        """
        return _FunctionThatRaises(f)

    @staticmethod
    def from_result_function(f: Callable[[str], Res[_Value]]) -> ParamType[_Value]:
        """
        Creates a parameter type from a function that returns a value or an error

        Args:
            f: Function that parses the string

        Returns:
            Parameter type
        """

        return _ResultFunction(f)

    @staticmethod
    def invalid() -> ParamType[_Value]:
        """
        Creates a parameter type that always return errors
        """

        def invalid_fun(s: str) -> NoReturn:
            raise RuntimeError("Invalid parameter type")

        return ParamType.from_function_that_raises(invalid_fun)

    @staticmethod
    def choices_str(
        values: Iterable[str],
        strip: bool = True,
        force_case: ForceCase = ForceCase.NO_CHANGE,
    ) -> ParamType[str]:
        """
        Creates a parameter type whose values are chosen from a set of strings

        Args:
            values: Set of values
            strip: Whether to strip whitespace before looking for choices

        Returns:
            Parameter type
        """

        return ParamType.choices({v: v for v in values}, strip, force_case)

    @staticmethod
    def choices(
        mapping: Mapping[str, _Value],
        strip: bool = True,
        force_case: ForceCase = ForceCase.NO_CHANGE,
        aliases: Mapping[str, _Value] = {},
    ) -> ParamType[_Value]:
        """
        Creates a parameter type whose strings correspond to keys in a dictionary

        Args:
            mapping: Dictionary mapping strings to values
            strip: Whether to strip whitespace before looking for keys
            force_case: Whether to normalize the case of the user string
            aliases: Additional mappings not shown in help

        Returns:
            Parameter type
        """
        return _Choices(mapping, strip, force_case, aliases)


@dataclass(frozen=True)
class _Choices(ParamType[_Value]):
    """
    Describes a multiple choice parameter type
    """

    mapping: Mapping[str, _Value]
    strip: bool
    force_case: ForceCase
    aliases: Mapping[str, _Value]

    def parse(self, arg: str) -> Res[_Value]:
        if self.strip:
            arg = arg.strip()
        if self.force_case is ForceCase.LOWER:
            arg = arg.lower()
        elif self.force_case is ForceCase.UPPER:
            arg = arg.upper()
        elif self.force_case is ForceCase.NO_CHANGE:
            pass
        else:
            assert_never(self.force_case)
        all_mappings = {**self.mapping, **self.aliases}
        if arg in all_mappings:
            return all_mappings[arg]
        else:
            msg = f"Value {arg} not in choices {','.join(self.mapping.keys())}"
            return Err.make(msg)

    def argparse_argument_kwargs(self) -> Mapping[str, Any]:
        return {"choices": self.mapping.keys(), "type": str}


@dataclass  # not frozen because mypy bug, please be responsible
class _FunctionThatRaises(ParamType[_Value]):
    """
    Wraps a function that may raise exceptions
    """

    # the optional is to make mypy happy
    fun: Callable[[str], _Value]  #: Callable function that may raise

    def parse(self, arg: str) -> Res[_Value]:
        try:
            f = self.fun
            assert f is not None
            return f(arg)
        except Exception as err:
            return Err.make(f"Error '{err}' in '{arg}'")


@dataclass  # not frozen because mypy bug, please be responsible
class _ResultFunction(ParamType[_Value]):
    """
    Wraps a function that returns a result
    """

    fun: Callable[[str], Res[_Value]]

    def parse(self, arg: str) -> Res[_Value]:
        return self.fun(arg)


@dataclass(frozen=True)
class _Parsy(ParamType[_Value]):
    """
    Wraps a parser from the parsy library
    """

    parser: parsy.Parser

    def parse(self, arg: str) -> Res[_Value]:
        res = (self.parser << parsy.eof)(arg, 0)  # Inspired by Parser.parse
        if res.status:
            return cast(_Value, res.value)
        else:
            if res.furthest is not None:
                return Err.make(
                    f"Parse error '{res.expected}' in '{arg}' at position '{res.furthest}'"
                )
            else:
                return Err.make(f"Parse error '{res.expected}' in '{arg}'")


@dataclass(frozen=True)
class _EmptyMeansNone(ParamType[Optional[_Item]]):
    """
    Wraps an existing parameter type so that "empty means none"
    """

    wrapped: ParamType[_Item]  #: Wrapped ParamType called if value is not empty
    strip: bool  #:  Whether to strip whitespace before testing for empty

    def parse(self, value: str) -> Res[Optional[_Item]]:
        if self.strip:
            value = value.strip()
        if not value:
            return None
        else:
            return self.wrapped.parse(value)


@dataclass(frozen=True)
class _Mapped(ParamType[_ReturnType], Generic[_Parameter, _ReturnType]):
    """
    Wraps an existing parser and applies a function to its successful result
    """

    wrapped: ParamType[_Parameter]  #: Wrapped parser
    f: Optional[Callable[[_Parameter], _ReturnType]]  #: Mapping function, made optional as a hack

    def parse(self, arg: str) -> Res[_ReturnType]:
        assert self.f is not None
        res: Res[_Parameter] = self.wrapped.parse(arg)
        if isinstance(res, Err):
            return res
        return self.f(res)


@dataclass(frozen=True)
class _SeparatedBy(ParamType[Sequence[_Item]]):
    """
    Parses values separated by a given separator
    """

    item: ParamType[_Item]  #: ParamType of individual items
    sep: str  #: Item separator
    strip: bool  #: Whether to strip whitespace around separated strings
    remove_empty: bool  #: Whether to prune empty strings

    def parse(self, arg: str) -> Res[Sequence[_Item]]:
        items: Iterable[str] = arg.split(self.sep)
        if self.strip:
            items = map(lambda s: s.strip(), items)
        if self.remove_empty:
            items = filter(None, items)
        res: Sequence[Res[_Item]] = [self.item.parse(s) for s in items]

        return collect_seq(res)


@dataclass  # TODO: (frozen=True) when mypy sorts out dataclasses with Callable fields
class _Validated(ParamType[_Value]):
    """
    Parses a value and validates it
    """

    wrapped: ParamType[_Value]
    predicate: Callable[[_Value], bool]
    msg: Callable[[str, _Value], str]

    def parse(self, arg: str) -> Res[_Value]:
        res = self.wrapped.parse(arg)
        if isinstance(res, Err):
            return res
        p = self.predicate
        if p(res):
            return res
        return Err.make(self.msg(arg, res))


#: Parses a path
path: ParamType[pathlib.Path] = ParamType.from_function_that_raises(lambda s: pathlib.Path(s))

#: Parses an integer
int_: ParamType[int] = ParamType.from_function_that_raises(lambda s: int(s))

#: Parses a word, stripping whitespace on the left and right
word: ParamType[str] = ParamType.from_function_that_raises(lambda s: s.strip())

#: Float parameter type
float_: ParamType[float] = ParamType.from_function_that_raises(lambda s: float(s))
bool_: ParamType[bool] = ParamType.choices(
    {"true": True, "false": False},
    force_case=ForceCase.LOWER,
    aliases={"t": True, "f": False, "1": True, "0": False},
)



__version__ = "8.0.0"

from . import types
from .arg import Derived, Expander, Param, Positional
from .config import Config
from .types import ParamType
from .userr import Err, Res

__all__ = [
    "Config",
    "Derived",
    "Err",
    "Expander",
    "Res",
    "Param",
    "ParamType",
    "Positional",
    "types",
]
